# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

## \file worldobject.py
## Contains the WorldObject class.

#from _core import WorldObject as _WorldObject
#from _core import _WorldObjectChildIterator
from . import _core
from .Interfaces import ISceneItem, ISceneItemContainer
from . import protocols
from . import globalscene
from .cgtypes import *

# WorldObject
class WorldObject(_core.WorldObject):
    """The base class for a world object.

    A world object is a scene item that has a position and orientation in
    space and that can usually be represented by 3D geometry.

    Attributes:

    - name
    - transform
    - pos
    - rot
    - scale
    - pivot
    - mass

    If you access an attribute that's actually stored in the geom then
    the attribute access is forwarded to the geom. This means you can
    access the geom slots through the world object (for example, if you
    attach a sphere geom to a world object, then the world object "has"
    an attribute 'radius').
    """

    protocols.advise(instancesProvide=[ISceneItem, ISceneItemContainer])
    
    def __init__(self,
                 name="object",
                 transform = None,
                 pos = None, rot = None, scale = None,
                 pivot = None,
                 offsetTransform = None,
                 parent = None,
                 mass = None,
                 material = None,
                 visible = True,
                 linearvel = None,
                 angularvel = None,
                 auto_insert = True):
        """Constructor.

        \param name (\c str) Object name
        \param transform (\c mat4) Initial transform
        \param pos (\c vec3) Initial position
        \param rot (\c mat3) Initial rotation
        \param scale (\c vec3) Initial scaling
        \param pivot (\c vec3) Initial pivot point (takes precedence over offsetTransform)
        \param offsetTransform (\c mat4) Initial offset transform
        \param parent (\c WorldObject or \c str) Parent object or None
        \param mass (\c float) Total mass
        \param material (\c Material) Material class (or a sequence of materials)
        \param visible (\c Bool) Visibility flag
        \param linearvel (\c vec3) Linear velocity
        \param angularvel (\c vec3) Angular velocity
        \param auto_insert (\c bool) If True, the object is inserted into the
                        scene automatically
        """

        _initWorldObject(self, baseClass=_core.WorldObject,
                         name=name,
                         transform=transform, pos=pos, rot=rot,
                         scale=scale, pivot=pivot,
                         offsetTransform=offsetTransform,
                         parent=parent,
                         mass=mass, material=material,
                         visible=visible,
                         linearvel=linearvel, angularvel=angularvel,
                         auto_insert=auto_insert)
        
#        if offsetTransform!=None:
#            self.setOffsetTransform(offsetTransform)
#        if pivot!=None:
#            self.pivot = pivot
#        if transform!=None:
#            self.transform = transform
#        if pos!=None:
#            self.pos = pos
#        if rot!=None:
#            self.rot = rot
#        if scale!=None:
#            self.scale = scale
#        if mass!=None:
#            self.mass = mass
#        if material!=None:
#            self.material = material

#        if auto_insert:
#            scene.getScene().insert(self)

    def protocols(self):
        return [ISceneItem, ISceneItemContainer]

    def __getattr__(self, name):
        if self.geom!=None and name[:2]!="__":
            if hasattr(self.geom, name):
                return getattr(self.geom, name)
            elif self.geom.hasSlot(name):
                return self.geom.slot(name).getValue()
#        if self.geom!=None and self.geom.hasSlot(name):
#            exec "res=self.geom.%s"%name
#            return res
        raise AttributeError('Object "%s" has no attribute "%s"'%(self.name, name))
    
    def __setattr__(self, name, val):
        if self.geom!=None and self.geom.hasSlot(name):
#            print "self.geom.%s=%s"%(name, val)
            exec("self.geom.%s=%s"%(name, val))
        else:
            _core.WorldObject.__setattr__(self, name, val)
        

# Common WorldObject initializations
def _initWorldObject(self, baseClass,
                     name, parent, transform=None,
                     pos=None, rot=None, scale=None,
                     pivot=None, offsetTransform=None,
                     mass=None, material=None, visible=True,
                     linearvel=None, angularvel=None,
                     auto_insert=True):
    """Helper function for usage in constructors.
    """
    if auto_insert:
        if parent is None:
            parent = globalscene.getScene().worldRoot()
        else:
            if type(parent) is str:
                parent = globalscene.getScene().worldObject(parent)
        name = parent.makeChildNameUnique(name)
    
    baseClass.__init__(self, name)
    
    if offsetTransform!=None:
        self.setOffsetTransform(offsetTransform)
    if pivot!=None:
        self.pivot = vec3(pivot)
    if transform!=None:
        self.transform = transform
    if pos!=None:
        self.pos = vec3(pos)
    if rot!=None:
        self.rot = mat3(rot)
    if scale!=None:
        self.scale = vec3(scale)
    if mass!=None:
        self.mass = mass
    if material!=None:
        try:
            # Check if material is a sequence or not. If it is not a
            # sequence the following line will raise an exception.
            len(material)
        except:
            material = [material]
        self.setNumMaterials(len(material))
        for i,mat in enumerate(material):
            self.setMaterial(mat, i)
    if linearvel!=None:
        self.linearvel = vec3(linearvel)
    if angularvel!=None:
        self.angularvel = vec3(angularvel)

    self.visible = visible

    if auto_insert:
        parent.addChild(self)
#        scene.getScene().insert(self)




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: material3ds.py,v 1.4 2005/08/28 19:42:43 mbaas Exp $

## \file material3ds.py
## Contains the Material3DS class.

import os.path, sys, shutil
import math
from . import protocols
from ._OpenGL.GL import *
from .Interfaces import *
from .slots import *
from . import lookat
from . import sl
from .material import Material
from .glmaterial import GLMaterial, GLTexture
from . import ribexport
from . import _Image as Image

# TextureMap3DS
class TextureMap3DS:
    """A texture map definition as it appears in 3DS files.
    """
    
    def __init__(self,
                 name,
                 flags = 0,
                 percent = 0.0,
                 blur = 0.0,
                 scale = (1.0, 1.0),
                 offset = (0,0),
                 rotation = 0,
                 tint1 = (0,0,0),
                 tint2 = (0,0,0),
                 tintr = (0,0,0),
                 tintg = (0,0,0),
                 tintb = (0,0,0)
                 ):
        self.name = name
        self.flags = flags
        self.percent = percent
        self.blur = blur
        self.scale = scale
        self.offset = offset
        self.rotation = rotation
        self.tint1 = tint1
        self.tint2 = tint2
        self.tintr = tintr
        self.tintg = tintg
        self.tintb = tintb

    def __str__(self):
        s = "TextureMap3DS:\n"
        s += '  Name    : "%s"\n'%self.name
        s += '  Flags   : %d\n'%self.flags
        s += '  Percent : %s\n'%self.percent
        s += '  Blur    : %s\n'%self.blur
        s += '  Scale   : %s, %s\n'%self.scale
        s += '  Offset  : %s, %s\n'%self.offset
        s += '  Rotation: %s\n'%self.rotation
        s += '  Tint1   : %s\n'%(self.tint1,)
        s += '  Tint2   : %s\n'%(self.tint2,)
        s += '  TintR   : %s\n'%(self.tintr,)
        s += '  TintG   : %s\n'%(self.tintg,)
        s += '  TintB   : %s'%(self.tintb,)
        return s


# Material3DS
class Material3DS(Material):
    """This class represents a material as it appears in 3DS files.

    """

    protocols.advise(instancesProvide=[ribexport.IMaterial])
    
#    protocols.advise(instancesProvide=[ISceneItem])

    def __init__(self,
                 name="Material3DS",
                 ambient = (0,0,0,0),
                 diffuse = (1.0, 1.0, 1.0, 1.0),
                 specular = (1.0, 1.0, 1.0, 1.0),
                 shininess = 1,
                 shin_strength = 0,
                 use_blur = 0,
                 transparency = 0.0,
                 falloff = 0,
                 additive = 0,
                 use_falloff = 0,
                 self_illum = False,
                 self_ilpct = 0.0,
                 shading = 0,
                 soften = 0,
                 face_map = 0,
                 two_sided = 0,
                 map_decal = 0,
                 use_wire = 0,
                 use_wire_abs = 0,
                 wire_size = 0,
                 density = 1.0,
                 texture1_map = None,
                 texture1_mask = None,
                 texture2_map = None,
                 texture2_mask = None,
                 opacity_map = None,
                 opacity_mask = None,
                 bump_map = None,
                 bump_mask = None,
                 specular_map = None,
                 specular_mask = None,
                 shininess_map = None,
                 shininess_mask = None,
                 self_illum_map = None,
                 self_illum_mask = None,
                 reflection_map = None,
                 reflection_mask = None,

                 bump_size = 1.0   # Extra parameter to control the bump map
                 ):

        Material.__init__(self, name=name, density=density)

        self.ambient = ambient
        self.diffuse = diffuse
        self.specular = specular
        self.shininess = shininess
        self.shin_strength = shin_strength
        self.transparency = transparency
        self.self_illum = self_illum
        self.self_ilpct = self_ilpct
        self.texture1_map = texture1_map
        self.texture1_mask = texture1_mask
        self.texture2_map = texture2_map
        self.texture2_mask = texture2_mask
        self.opacity_map = opacity_map
        self.opacity_mask = opacity_mask
        self.bump_map = bump_map
        self.bump_mask = bump_mask
        self.specular_map = specular_map
        self.specular_mask = specular_mask
        self.shininess_map = shininess_map
        self.shininess_mask = shininess_mask
        self.self_illum_map = self_illum_map
        self.self_illum_mask = self_illum_mask
        self.reflection_map = reflection_map
        self.reflection_mask = reflection_mask
        self.bump_size = bump_size

        if texture1_map==None:
            self._gltexture = None
            glambient = ambient
            gldiffuse = diffuse
        else:
            map = texture1_map
            T = mat4(1,0,0,-map.offset[0]-0.5, 0,-1,0,0.5-map.offset[1], 0,0,1,0, 0,0,0,1)
            a = sl.radians(map.rotation)
            ca = math.cos(a)
            sa = math.sin(a)
            R = mat4(ca,-sa,0,0, sa,ca,0,0, 0,0,1,0, 0,0,0,1)
            S = mat4(map.scale[0],0,0,0.5, 0,map.scale[1],0,0.5, 0,0,1,0, 0,0,0,1)
            self._gltexture = GLTexture(imagename = map.name,
                                        mode = GL_MODULATE,
#                                        transform = mat4().scaling(vec3(1,-1,1))
                                        transform = S*R*T
                                        )
            glambient = map.percent*vec4(1,1,1,1) + (1-map.percent)*vec4(ambient)
            gldiffuse = map.percent*vec4(1,1,1,1) + (1-map.percent)*vec4(diffuse)

        self._glmaterial = GLMaterial(ambient = glambient,
                                      diffuse = gldiffuse,
                                      specular = shin_strength*specular,
                                      shininess = 25*shininess,
                                      texture = self._gltexture
                                      )

##        print >>sys.stderr, "---",name,"---"
##        print >>sys.stderr, "ambient",self.ambient
##        print >>sys.stderr, "diffuse",self.diffuse
##        print >>sys.stderr, "Map1:"
##        print >>sys.stderr, self.texture1_map
##        print >>sys.stderr, "Map2:"
##        print >>sys.stderr, self.texture2_map
##        print >>sys.stderr, "Opacity:"
##        print >>sys.stderr, self.opacity_map
##        print >>sys.stderr, "Bump:"
##        print >>sys.stderr, self.bump_map
##        print >>sys.stderr, "Specular:"
##        print >>sys.stderr, self.specular_map
##        print >>sys.stderr, "Shininess:"
##        print >>sys.stderr, self.shininess_map
##        print >>sys.stderr, "Self illum:"
##        print >>sys.stderr, self.self_illum_map
##        print >>sys.stderr, "Reflection:"
##        print >>sys.stderr, self.reflection_map
##        print >>sys.stderr, "shininess",shininess
##        print >>sys.stderr, "shin_strength",shin_strength
##        print >>sys.stderr, "transparency",transparency
##        print >>sys.stderr, "falloff",falloff
##        print >>sys.stderr, "use_falloff",use_falloff
##        print >>sys.stderr, "use_blur", use_blur
##        print >>sys.stderr, "additive",additive
##        print >>sys.stderr, "shading",shading
##        print >>sys.stderr, "soften",soften
##        print >>sys.stderr, "self_illum", self_illum
##        print >>sys.stderr, "face_map", face_map
##        print >>sys.stderr, "two_sided", two_sided
##        print >>sys.stderr, "map_decal", map_decal
##        print >>sys.stderr, "use_wire", use_wire
##        print >>sys.stderr, "use_wire_abs", use_wire_abs
##        print >>sys.stderr, "wire_size", wire_size

    def applyGL(self):
        self._glmaterial.applyGL()

    def usesBlending(self):
        return self._glmaterial.usesBlending()

        
    def createPasses(self):
        """Returns a list of RenderPass objects."""
        texdefs = []
        texdefs += self._createTexDef(self.texture1_map)
        texdefs += self._createTexDef(self.opacity_map)
        texdefs += self._createTexDef(self.bump_map)
        texdefs += self._createTexDef(self.specular_map)
        texdefs += self._createTexDef(self.shininess_map)
        texdefs += self._createTexDef(self.self_illum_map)
        texdefs += self._createTexDef(self.reflection_map)

        if texdefs!=[]:
            return [ribexport.TexPass(maps=texdefs)]
        else:
            return []


    def preProcess(self, exporter):
        """Preprocessing method.

        This method is called before the image is rendered and can be used
        to create or copy image maps.
        """
        pass
#        self._copyImageMap(self.texture1_map, exporter)                 
#        self._copyImageMap(self.opacity_map, exporter)
#        self._copyImageMap(self.bump_map, exporter)
#        self._copyImageMap(self.specular_map, exporter)
#        self._copyImageMap(self.shininess_map, exporter)
#        self._copyImageMap(self.self_illum_map, exporter)
#        self._copyImageMap(self.reflection_map, exporter)
            

    def color(self):
        """Return the color for the RiColor() call or None.
        """
        return self.diffuse

    def opacity(self):
        """Return the opacity for the RiOpacity() call or None.
        """
        a = 1.0-self.transparency
        return (a,a,a)

    def surfaceShaderName(self):
        """Returns the name of the corresponding surface shader or None.
        """
        return "mat3ds"

    def surfaceShaderSource(self):
        """Returns surface shader source code as a string or None.

        If the return value is None, then shaderName() must return
        the name of the shader to use.
        """
        return """// 3DS material shader
        
surface $SHADERNAME(color ambient_col = color "rgb" (0, 0, 0);
           color diffuse_col = color "rgb" (1, 1, 1);
           color specular_col = color "rgb" (1, 1, 1);
           float shininess = 0.0;
           float shin_strength = 0.0;
           float self_ilpct = 0.0;
           // Texture map 1
           string texture1_map = "";
           float t1_uscale = 1.0;
           float t1_vscale = 1.0;
           float t1_uoffset = 0;
           float t1_voffset = 0;
           float t1_rotation = 0;
           float t1_blur = 0.0;
           // Specular map
           string specular_map = "";
           float sp_uscale = 1.0;
           float sp_vscale = 1.0;
           float sp_uoffset = 0;
           float sp_voffset = 0;
           float sp_rotation = 0;
           float sp_blur = 0.0;
           // Shininess map
           string shininess_map = "";
           float sh_uscale = 1.0;
           float sh_vscale = 1.0;
           float sh_uoffset = 0;
           float sh_voffset = 0;
           float sh_rotation = 0;
           float sh_blur = 0.0;
           // Opacity map
           string opacity_map = "";
           float op_uscale = 1.0;
           float op_vscale = 1.0;
           float op_uoffset = 0;
           float op_voffset = 0;
           float op_rotation = 0;
           float op_blur = 0.0;
           // Self illumination map
           string self_illum_map = "";
           float si_uscale = 1.0;
           float si_vscale = 1.0;
           float si_uoffset = 0;
           float si_voffset = 0;
           float si_rotation = 0;
           float si_blur = 0.0;
           // Reflection map
           string reflection_map = "";
           float refl_uscale = 1.0;
           float refl_vscale = 1.0;
           float refl_uoffset = 0;
           float refl_voffset = 0;
           float refl_rotation = 0;
           float refl_percent = 0.0;
           float refl_blur = 0.0;
           varying point Pref = point(0,0,0);
           )
{
  BAKE_BEGIN
  normal Nf = BAKE_NORMAL(N);
  color C_diffuse = diffuse_col;
  color C_specular = specular_col;
  color C_opacity = Os;
  color C_refl = 0;
  float final_shininess = shininess;
  float final_self_ilpct = self_ilpct;
  float s0, t0, ss, tt, a, ca, sa;
  color col;

  if (texture1_map!="")
  {
    a = radians(t1_rotation);
    ca = cos(a);
    sa = sin(a);
    s0 = s-t1_uoffset-0.5;
    t0 = 0.5-t-t1_voffset;
    ss = ca*s0 - sa*t0;
    tt = sa*s0 + ca*t0;
    ss = t1_uscale*ss+0.5;
    tt = t1_vscale*tt+0.5;
    C_diffuse = texture(texture1_map, ss, tt, "blur", t1_blur);
  }

  if (specular_map!="")
  {
    a = radians(sp_rotation);
    ca = cos(a);
    sa = sin(a);
    s0 = s-sp_uoffset-0.5;
    t0 = 0.5-t-sp_voffset;
    ss = ca*s0 - sa*t0;
    tt = sa*s0 + ca*t0;
    ss = sp_uscale*ss+0.5;
    tt = sp_vscale*tt+0.5;
    C_specular = texture(specular_map, ss, tt, "blur", sp_blur);
  }

  if (shininess_map!="")
  {
    a = radians(sh_rotation);
    ca = cos(a);
    sa = sin(a);
    s0 = s-sh_uoffset-0.5;
    t0 = 0.5-t-sh_voffset;
    ss = ca*s0 - sa*t0;
    tt = sa*s0 + ca*t0;
    ss = sh_uscale*ss+0.5;
    tt = sh_vscale*tt+0.5;
    col = texture(shininess_map, ss, tt, "blur", sh_blur);
    final_shininess = (comp(col,0)+comp(col,1)+comp(col,2))/3;
  }

  if (opacity_map!="")
  {
    a = radians(op_rotation);
    ca = cos(a);
    sa = sin(a);
    s0 = s-op_uoffset-0.5;
    t0 = 0.5-t-op_voffset;
    ss = ca*s0 - sa*t0;
    tt = sa*s0 + ca*t0;
    ss = op_uscale*ss+0.5;
    tt = op_vscale*tt+0.5;
    float op = texture(opacity_map, ss, tt, "blur", op_blur);
    C_opacity = color "rgb" (op, op, op);
  }

  if (self_illum_map!="")
  {
    a = radians(si_rotation);
    ca = cos(a);
    sa = sin(a);
    s0 = s-si_uoffset-0.5;
    t0 = 0.5-t-si_voffset;
    ss = ca*s0 - sa*t0;
    tt = sa*s0 + ca*t0;
    ss = si_uscale*ss+0.5;
    tt = si_vscale*tt+0.5;
    col = texture(self_illum_map, ss, tt, "blur", si_blur);
    final_self_ilpct = (comp(col,0)+comp(col,1)+comp(col,2))/3;
  }

  if (reflection_map!="")
  {
    vector R = normalize(vtransform("world", reflect(I, Nf)));
    float lat = acos(zcomp(R));
    float long = acos(ycomp(R)/sin(lat));
    lat = lat/PI;
    long = long/(2*PI);
    if (xcomp(R)>0)
      long = -long;
    C_refl = texture(reflection_map, long, lat, "blur", refl_blur);
  }

  Ci = C_diffuse*diffuse(Nf) +
       shin_strength*C_specular*phong(Nf, -normalize(I), 25*final_shininess) +
       refl_percent*C_refl;
  Ci = mix(Ci, C_diffuse, final_self_ilpct);
  Oi = C_opacity;
  Ci *= Oi;
  BAKE_END
}        
        """

    def surfaceShaderParams(self, passes):
        """Return a dictionary with shader parameters and their values."""
        a = self.ambient
        d = self.diffuse
        s = self.specular
        res = {"uniform color ambient_col" : (a[0], a[1], a[2]),
                "uniform color diffuse_col" : (d[0], d[1], d[2]),
                "uniform color specular_col" : (s[0], s[1], s[2]),
                "uniform float shininess" : self.shininess,
                "uniform float shin_strength" : self.shin_strength,
                "uniform float self_ilpct" : self.self_ilpct
                }

        if self.texture1_map!=None:
            texname = os.path.basename(self.texture1_map.name)
            name, ext = os.path.splitext(texname)
            res["uniform string texture1_map"] = name+".tex"
            res["uniform float t1_uscale"] = self.texture1_map.scale[0]
            res["uniform float t1_vscale"] = self.texture1_map.scale[1]
            res["uniform float t1_uoffset"] = self.texture1_map.offset[0]
            res["uniform float t1_voffset"] = self.texture1_map.offset[1]
            res["uniform float t1_rotation"] = self.texture1_map.rotation

        if self.specular_map!=None:
            texname = os.path.basename(self.specular_map.name)
            name, ext = os.path.splitext(texname)
            res["uniform string specular_map"] = name+".tex"
            res["uniform float sp_uscale"] = self.specular_map.scale[0]
            res["uniform float sp_vscale"] = self.specular_map.scale[1]
            res["uniform float sp_uoffset"] = self.specular_map.offset[0]
            res["uniform float sp_voffset"] = self.specular_map.offset[1]
            res["uniform float sp_rotation"] = self.specular_map.rotation

        if self.shininess_map!=None:
            texname = os.path.basename(self.shininess_map.name)
            name, ext = os.path.splitext(texname)
            res["uniform string shininess_map"] = name+".tex"
            res["uniform float sh_uscale"] = self.shininess_map.scale[0]
            res["uniform float sh_vscale"] = self.shininess_map.scale[1]
            res["uniform float sh_uoffset"] = self.shininess_map.offset[0]
            res["uniform float sh_voffset"] = self.shininess_map.offset[1]
            res["uniform float sh_rotation"] = self.shininess_map.rotation

        if self.opacity_map!=None:
            texname = os.path.basename(self.opacity_map.name)
            name, ext = os.path.splitext(texname)
            res["uniform string opacity_map"] = name+".tex"
            res["uniform float op_uscale"] = self.opacity_map.scale[0]
            res["uniform float op_vscale"] = self.opacity_map.scale[1]
            res["uniform float op_uoffset"] = self.opacity_map.offset[0]
            res["uniform float op_voffset"] = self.opacity_map.offset[1]
            res["uniform float op_rotation"] = self.opacity_map.rotation

        if self.self_illum_map!=None:
            texname = os.path.basename(self.self_illum_map.name)
            name, ext = os.path.splitext(texname)
            res["uniform string self_illum_map"] = name+".tex"
            res["uniform float si_uscale"] = self.self_illum_map.scale[0]
            res["uniform float si_vscale"] = self.self_illum_map.scale[1]
            res["uniform float si_uoffset"] = self.self_illum_map.offset[0]
            res["uniform float si_voffset"] = self.self_illum_map.offset[1]
            res["uniform float si_rotation"] = self.self_illum_map.rotation

        if self.reflection_map!=None:
            texname = os.path.basename(self.reflection_map.name)
            name, ext = os.path.splitext(texname)
            res["uniform string reflection_map"] = name+".tex"
            res["uniform float refl_uscale"] = self.reflection_map.scale[0]
            res["uniform float refl_vscale"] = self.reflection_map.scale[1]
            res["uniform float refl_uoffset"] = self.reflection_map.offset[0]
            res["uniform float refl_voffset"] = self.reflection_map.offset[1]
            res["uniform float refl_rotation"] = self.reflection_map.rotation
            res["uniform float refl_percent"] = self.reflection_map.percent
            res["uniform float refl_blur"] = self.reflection_map.blur

        return res

    def surfaceShaderTransform(self):
        return mat4(1)

    def displacementShaderName(self):
        if self.bump_map==None:
            return None
        
        return "mat3ds_bump"

    
    def displacementShaderSource(self):
        if self.bump_map==None:
            return None
        
        return """// 3DS material shader (bump)
        
displacement $SHADERNAME(
           // Bump map
           string bump_map = "";
           float uscale = 1.0;
           float vscale = 1.0;
           float uoffset = 0;
           float voffset = 0;
           float rotation = 0;
           float blur = 0;
           float bump_size = 1.0;
           )
{
  float s0, t0, ss, tt, a, ca, sa;
  color bc;
  float amount = 0.0;

  if (bump_map!="")
  {
    a = radians(rotation);
    ca = cos(a);
    sa = sin(a);
    s0 = s-uoffset-0.5;
    t0 = 0.5-t-voffset;
    ss = ca*s0 - sa*t0;
    tt = sa*s0 + ca*t0;
    ss = uscale*ss+0.5;
    tt = vscale*tt+0.5;
    bc = texture(bump_map, ss, tt, "blur", blur);
    amount = (comp(bc,0)+comp(bc,1)+comp(bc,2))/3;
  }

  P = P - bump_size*amount*normalize(N);
  N = calculatenormal(P);
}
        """
    
    def displacementShaderParams(self, passes):
        if self.bump_map==None:
            return {}

        res = {}
        texname = os.path.basename(self.bump_map.name)
        name, ext = os.path.splitext(texname)
        res["uniform string bump_map"] = name+".tex"
        res["uniform float uscale"] = self.bump_map.scale[0]
        res["uniform float vscale"] = self.bump_map.scale[1]
        res["uniform float uoffset"] = self.bump_map.offset[0]
        res["uniform float voffset"] = self.bump_map.offset[1]
        res["uniform float rotation"] = self.bump_map.rotation
        res["uniform float blur"] = self.bump_map.blur
        res["uniform float bump_size"] = self.bump_size
        return res

    def displacementBound(self):
        if self.bump_map==None:
            return "current", 0
        else:
            return "current", self.bump_size

    def displacementShaderTransform(self):
        return mat4(1)
	
    def interiorShaderName(self):
        return None
    
    def interiorShaderSource(self):
        return None
    
    def interiorShaderParams(self, passes):
        return {}

    def interiorShaderTransform(self):
        return mat4(1)


##    def _copyImageMap(self, texmap, exporter):
##        """Copy the texture map image into the map folder.

##        \param texmap (\c TextureMap3DS) Texture map object or None
##        \param exporter Exporter instance
##        """
##        return
##        if texmap==None:
##            return

##        exporter.checkMapPath()
##        texname = os.path.basename(texmap.name)
##        name, ext = os.path.splitext(texname)
##        if ext.lower()!=".tif":
##            print 'Converting "%s"'%texmap.name
##            tifname = os.path.join(exporter.map_path, name+".tif")
##            # Read original map
##            try:
##                img = Image.open(texmap.name)
##            except IOError, e:
##                print e
##                return
##            # Save map as TIF file
##            img.save(tifname)
##        else:
##            print 'Copying "%s"'%texmap.name
##            shutil.copyfile(texmap.name, os.path.join(exporter.map_path, texmap.name))
         

    def _createTexDef(self, texmap):
        """Create texture definition for the TexPass.

        The method returns a list that is either empty or contains
        one texture definition tuple.

        \param texmap (\c TextureMap3DS) Texture map object or None
        \return An empty list or a list containing one texture def
        """
        if texmap==None:
            return []
        
#        texname = os.path.basename(texmap.name)
        texname = texmap.name
#        texname, ext = os.path.splitext(texname)
        blur = texmap.blur+1.0
        texdef = (texname, #+".tif",
                  "periodic", "periodic",
                  "gaussian", blur, blur, {})
        return [texdef]
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: joystick.py,v 1.1.1.1 2004/12/12 14:31:05 mbaas Exp $

## \file joystick.py
## Contains the Joystick class.

from .eventmanager import eventManager
from .events import *

# Joystick
class Joystick:
    """Generic joystick class.

    Attributes:

    - id
    - name
    - numaxes
    - numhats
    - numballs
    - numbuttons

    """
    
    def __init__(self, id, name,
                 numaxes=0, numhats=0, numballs=0, numbuttons=0):
        """Constructor.

        \param id (\c int) Device ID (0,1,2...)
        \param name (\c str) The name of the joystick device
        \param numaxes (\c int) The number of axes this joystick has
        \param numhats (\c int) The number of hats this joystick has
        \param numballs (\c int) The number of balls this joystick has
        \param numbuttons (\c int) The number of buttons this joystick has
        """

        self.id = id
        self.name = name
        self.numaxes = numaxes
        self.numhats = numhats
        self.numballs = numballs
        self.numbuttons = numbuttons

        # Current uncalibrated axis values
        self.axis = numaxes*[0.0]
        # Current hat values
        self.hat = numhats*[(0,0)]
        # Current ball values
        self.ball = numballs*[0.0]
        # Current button states
        self.button = numbuttons*[False]

        # Minimum axis values
        self.axis_min = numaxes*[999.0]
        # Maximum axis values
        self.axis_max = numaxes*[-999.0]
        # Middle axis values
        self.axis_mid = numaxes*[0.0]

    def __str__(self):
        return '<Joystick id:%d name:"%s" axes:%d hats:%d balls:%d buttons:%d>'%(self.id, self.name, self.numaxes, self.numhats, self.numballs, self.numbuttons)

    # getAxis
    def getAxis(self, axis):
        """Return an axis value.

        \return Calibrated axis value (\c float).
        """
        if axis>=self.numaxes:
            return 0.0
            
        return self._calibratedAxis(axis, self.axis[axis])

    # getHat
    def getHat(self, hat):
        """Return a hat value.

        \return A tuple (x, y) (\c int, \c int)
        """
        if hat>=self.numhats:
            return 0,0
        
        return self.hat[hat]

    # getBall
    def getBall(self, ball):
        """Return a ball value.

        \return Ball value (\c float)
        """
        if ball>=self.numballs:
            return 0.0
        
        return self.ball[ball]

    # getButton
    def getButton(self, button):
        """Return a button state.

        \return Button state (\c bool)
        """
        if button>=self.numbuttons:
            return False
        
        return self.button[button]

    # setAxis
    def setAxis(self, axis, value):
        """Set a new value for a particular axis.

        The value is the raw (uncalibrated) value from the joystick.

        This method sends a JOYSTICK_AXIS event which contains the
        \em calibrated axis value.

        \param axis (\c int) Axis index
        \param value (\c float) Uncalibrated axis value
        
        \todo Sollte diese Methode setRawAxis() heissen?
        """
        self.axis[axis] = value

        # Adjust min/max values
        if value<self.axis_min[axis]:
            self.axis_min[axis] = value
        if value>self.axis_max[axis]:
            self.axis_max[axis] = value
        
        e = JoystickAxisEvent(self.id, axis, self._calibratedAxis(axis, value))
        eventManager().event(JOYSTICK_AXIS, e)

    # setHat
    def setHat(self, hat, x, y):
        """Set a new value for a particular hat.

        This method sends a JOYSTICK_HAT event.

        \param hat (\c int) Hat index
        \param x (\c int) X value
        \param y (\c int) Y value        
        """
        self.hat[hat]=(x,y)
        e = JoystickHatEvent(self.id, hat, x, y)
        eventManager().event(JOYSTICK_HAT, e)

    # setBall
    def setBall(self, ball, value):
        """Set a new value for a particular ball.

        This method sends a JOYSTICK_BALL event.

        \param ball (\c int) Ball index
        \param value (\c float) Ball value
        """
        self.ball[ball] = value
        e = JoystickBallEvent(self.id, ball, value)
        eventManager().event(JOYSTICK_BALL, e)

    # setButton
    def setButton(self, button, value):
        """Set the state of a particular button.

        This method sends either a JOYSTICK_BUTTON_DOWN or
        JOYSTICK_BUTTON_UP event.

        \param button (\c int) Button index
        \param value (\c bool) State of the button (True = pressed)
        """
        self.button[button] = value
        e = JoystickButtonEvent(self.id, button)
        if value:
            eventManager().event(JOYSTICK_BUTTON_DOWN, e)                
        else:
            eventManager().event(JOYSTICK_BUTTON_UP, e)

    # setAxisMiddle
    def setAxisMiddle(self, axis=None, mid=None):
        """Set the middle position of an axis or all axes.

        Sets the joystick position that will be mapped to the value 0.0.
        This means, if the joystick outputs the (uncalibrated) value
        \a mid, then the calibrated value will just be 0.0.
        
        \param axis (\c int) Axis index (None = all axes)
        \param mid (\c float) Middle value (None = take current value)
        """
        if axis==None:
            axes = list(range(self.numaxes))
        else:
            axes = [axis]

        for i in axes:
            if mid==None:
                mid = self.axis[i]

            self.axis_mid[i] = mid

    ## protected:

    def _calibratedAxis(self, axis, rawvalue):
        """Return the calibrated axis value.

        \param axis (\c int) Axis index
        \param rawvalue (\c float) Uncalibrated axis value
        \return Calibrated axis value (\c float).
        """
        amid = self.axis_mid[axis]
        amin = self.axis_min[axis]
        amax = self.axis_max[axis]
        
        if rawvalue<amid:
            d = amid-amin
            if d>0:
                return (rawvalue-amid)/d
            else:
                return 0.0
        else:
            d = amax-amid
            if d>0:
                return (rawvalue-amid)/d
            else:
                return 0.0
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

"""This module contains the MBReader base class to parse Maya binary files (~IFF files).
"""

from . import mayaiff

class MBReader(mayaiff.IFFReader):
    """Read Maya binary files and call an appropriate handler for each chunk.
    
    This is the base class for a .mb reader class. Derived classes can implement
    the chunk handler methods onBeginGroup(), onEndGroup() and onDataChunk().
    These handlers receive a Chunk (or GroupChunk) object as input that
    contains information about the current chunk and that can also be used
    to read the actual chunk data.
    """
    
    def __init__(self):
        mayaiff.IFFReader.__init__(self, "Maya")
            
            
if __name__=="__main__":
    
    import sys
    
    class MBDumper(MBReader):
        def onBeginGroup(self, chunk):
            print(("GRP BEGIN %s"%chunk))
        
        def onEndGroup(self, chunk):
            print(("GRP END %s"%chunk))
        
        def onDataChunk(self, chunk):
            print(("CHUNK %s"%chunk))
            
    rd = MBDumper()
    rd.read(open(sys.argv[1], "rb"))
    


# coding: latin1
# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: mat3.py,v 1.2 2005/08/17 19:38:29 mbaas Exp $

import math, copy
from .vec3 import vec3 as _vec3

# [  0   1   2 ]
# [  3   4   5 ]
# [  6   7   8 ]

# Comparison threshold
_epsilon = 1E-12


# mat3
class mat3:
    """Matrix class (3x3).

    This class represents a 3x3 matrix that can be used to store
    linear transformations.
    """

    def __init__(self, *args):
        """Constructor.

        There are several possibilities how to initialize a matrix,
        depending on the number of arguments you provide to the constructor.

        - 0 arguments: Every component is zero.
        - 1 number argument: The diagonal is initialized to that number,
          all the other elements are zero.
        - 1 sequence argument: The elements are initialized with the numbers
          in the sequence (the sequence must contain 9 numbers).
        - 1 mat3 argument: The matrix is copied.
        - 3 sequence arguments: The columns are initialized with the
          respective sequence (each sequence must contain 3 numbers).
        - 9 number arguments: The matrix is initialized with those values
          (row-major order).
        """

        # No arguments
        if len(args)==0:
            self.mlist = 9*[0.0]

        # 1 argument (list, scalar or mat3)
        elif len(args)==1:
            T = type(args[0])
            # Scalar
            if T in [float, int]:
                f = float(args[0])
                self.mlist = [f,0.0,0.0,
                              0.0,f,0.0,
                              0.0,0.0,f]
            # mat3
            elif isinstance(args[0], mat3):
                self.mlist = copy.copy(args[0].mlist)
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                self.mlist=list([float(x) for x in s])
            else:
                self.mlist = mat3(*args[0]).mlist

        # 3 arguments (sequences)
        elif len(args)==3:
            a,b,c=args
            self.mlist = [a[0], b[0], c[0],
                          a[1], b[1], c[1],
                          a[2], b[2], c[2]]
            self.mlist = list([float(x) for x in self.mlist])

        # 9 arguments
        elif len(args)==9:
            self.mlist = list([float(x) for x in args])

        else:
            raise TypeError("mat3() arg can't be converted to mat3")

        # Check if there are really 9 elements in the list
        if len(self.mlist)!=9:
            raise TypeError("mat4(): Wrong number of matrix elements (%s instead of 9)"%(len(self.mlist)))


    def __repr__(self):
        return 'mat3(%s)'%(repr(self.mlist)[1:-1])

    def __str__(self):
        fmt="%9.4f"
        m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
        return ('['+fmt%m11+', '+fmt%m12+', '+fmt%m13+']\n'+
                '['+fmt%m21+', '+fmt%m22+', '+fmt%m23+']\n'+
                '['+fmt%m31+', '+fmt%m32+', '+fmt%m33+']')

    def __eq__(self, other):
        """== operator"""
        global _epsilon
        if isinstance(other, mat3):
#            return self.mlist==other.mlist
            lst = list([a for a in zip(self.mlist, other.mlist) if abs(a[0]-a[1])>_epsilon])
            return len(lst)==0
        else:
            return False

    def __ne__(self, other):
        """!= operator"""
        return not (self==other)


    def __add__(self, other):
        if isinstance(other, mat3):
            return mat3(list(map(lambda x,y: x+y, self.mlist, other.mlist)))
        else:
            raise TypeError("unsupported operand type for +")

    def __sub__(self, other):
        if isinstance(other, mat3):
            return mat3(list(map(lambda x,y: x-y, self.mlist, other.mlist)))
        else:
            raise TypeError("unsupported operand type for -")

    def __mul__(self, other):
        T = type(other)
        # mat3*scalar
        if T in [float, int]:
            return mat3(list(map(lambda x,other=other: x*other, self.mlist)))
        # mat3*vec3
        if isinstance(other, _vec3):
            m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
            return _vec3(m11*other.x + m12*other.y + m13*other.z, 
                         m21*other.x + m22*other.y + m23*other.z, 
                         m31*other.x + m32*other.y + m33*other.z)            
        # mat3*mat3
        if isinstance(other, mat3):
            m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
            n11,n12,n13,n21,n22,n23,n31,n32,n33 = other.mlist
            return mat3( m11*n11+m12*n21+m13*n31,
                         m11*n12+m12*n22+m13*n32,
                         m11*n13+m12*n23+m13*n33,

                         m21*n11+m22*n21+m23*n31,
                         m21*n12+m22*n22+m23*n32,
                         m21*n13+m22*n23+m23*n33,

                         m31*n11+m32*n21+m33*n31,
                         m31*n12+m32*n22+m33*n32,
                         m31*n13+m32*n23+m33*n33)
        # unsupported
        else:
            raise TypeError("unsupported operand type for *")

    def __rmul__(self, other):
        T = type(other)
        # scalar*mat3
        if T in [float, int]:
            return mat3(list(map(lambda x,other=other: other*x, self.mlist)))
        # vec3*mat3
        if isinstance(other, _vec3):
            m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
            return _vec3(other.x*m11 + other.y*m21 + other.z*m31, 
                         other.x*m12 + other.y*m22 + other.z*m32, 
                         other.x*m13 + other.y*m23 + other.z*m33)
        # mat3*mat3
        if isinstance(other, mat3):
            return self.__mul__(other)
        # unsupported
        else:
            raise TypeError("unsupported operand type for *")


    def __truediv__(self, other):
        T = type(other)
        # mat3/scalar
        if T in [float, int]:
            return mat3(list(map(lambda x,other=other: x/other, self.mlist)))
        # unsupported
        else:
            raise TypeError("unsupported operand type for /")

    def __mod__(self, other):
        T = type(other)
        # mat3%scalar
        if T in [float, int]:
            return mat3(list(map(lambda x,other=other: x%other, self.mlist)))
        # mat3%mat3
        if isinstance(other, mat3):
            return mat3([a[0]%a[1] for a in zip(self.mlist, other.mlist)])
        # unsupported
        else:
            raise TypeError("unsupported operand type for %")


    def __neg__(self):
        return mat3([-x for x in self.mlist])

    def __pos__(self):
        return mat3([+x for x in self.mlist])


    def __len__(self):
        return 3

    def __getitem__(self, key):
        """Return a column or an individual element."""
        if type(key)==int:
            if   key==0:
                return _vec3(self.mlist[0],self.mlist[3],self.mlist[6])
            elif key==1:
                return _vec3(self.mlist[1],self.mlist[4],self.mlist[7])
            elif key==2:
                return _vec3(self.mlist[2],self.mlist[5],self.mlist[8])
            else:
                raise IndexError("index out of range")
        elif type(key) is tuple:
            i,j=key
            if i<0 or i>2 or j<0 or j>2:
                raise IndexError("index out of range")
            return self.mlist[i*3+j]
        else:
            raise TypeError("index must be integer or 2-tuple")

    def __setitem__(self, key, value):
        """Set a column or an individual element."""
        if type(key)==int:
            value = [float(x) for x in value]
            if   key==0: self.mlist[0],self.mlist[3],self.mlist[6] = value
            elif key==1: self.mlist[1],self.mlist[4],self.mlist[7] = value
            elif key==2: self.mlist[2],self.mlist[5],self.mlist[8] = value
            else:
                raise IndexError("index out of range")
        elif type(key) is tuple:
            i,j=key
            if i<0 or i>2 or j<0 or j>2:
                raise IndexError("index out of range")
            self.mlist[i*3+j] = float(value)
        else:
            raise TypeError("index must be integer or 2-tuple")

    def getRow(self, index):
        """Return a row (as vec3)."""
        if type(index)==int:
            if index==0:
                return _vec3(self.mlist[0], self.mlist[1], self.mlist[2])
            elif index==1:
                return _vec3(self.mlist[3], self.mlist[4], self.mlist[5])
            elif index==2:
                return _vec3(self.mlist[6], self.mlist[7], self.mlist[8])
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be an integer")

    def setRow(self, index, value):
        """Set a row (as vec3)."""
        if type(index)==int:
            value = [float(x) for x in value]
            if index==0: self.mlist[0], self.mlist[1], self.mlist[2] = value
            elif index==1: self.mlist[3], self.mlist[4], self.mlist[5] = value
            elif index==2: self.mlist[6], self.mlist[7], self.mlist[8] = value
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be an integer")

    def getColumn(self, index):
        """Return a column (as vec3)."""
        if type(index)==int:
            if index==0:
                return _vec3(self.mlist[0], self.mlist[3], self.mlist[6])
            elif index==1:
                return _vec3(self.mlist[1], self.mlist[4], self.mlist[7])
            elif index==2:
                return _vec3(self.mlist[2], self.mlist[5], self.mlist[8])
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be an integer")

    def setColumn(self, index, value):
        """Set a column."""
        if type(index)==int:
            value = [float(x) for x in value]
            if index==0: self.mlist[0], self.mlist[3], self.mlist[6] = value
            elif index==1: self.mlist[1], self.mlist[4], self.mlist[7] = value
            elif index==2: self.mlist[2], self.mlist[5], self.mlist[8] = value
            else:
                raise IndexError("index out of range")
        else:
            raise TypeError("index must be an integer")

    def getDiag(self):
        """Return the diagonal."""
        return _vec3(self.mlist[0], self.mlist[4], self.mlist[8])

    def setDiag(self, value):
        """Set diagonal."""
        a,b,c = value
        self.mlist[0] = float(a)
        self.mlist[4] = float(b)
        self.mlist[8] = float(c)

    def toList(self, rowmajor=0):
        """Create a list containing the matrix elements.

        By default the list is in column-major order. If you set the
        optional argument rowmajor to 1, you'll get the list in row-major
        order.
        """
        if rowmajor:
            return copy.copy(self.mlist)
        else:
            return self.transpose().mlist

    def identity():
        """Return the identity matrix."""
        return mat3(1.0, 0.0, 0.0,
                    0.0, 1.0, 0.0,
                    0.0, 0.0, 1.0)
    identity = staticmethod(identity)

    def transpose(self):
        """Return the transposed matrix."""
        m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
        return mat3(m11,m21,m31,
                    m12,m22,m32,
                    m13,m23,m33)

    def determinant(self):
        """Return determinant."""
        m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
        return m11*m22*m33+ \
               m12*m23*m31+ \
               m13*m21*m32- \
               m31*m22*m13- \
               m32*m23*m11- \
               m33*m21*m12

    def inverse(self):
        """Return inverse matrix."""
        m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist
        d = 1.0/self.determinant()
        return mat3( m22*m33-m23*m32, m32*m13-m12*m33, m12*m23-m22*m13,
                     m23*m31-m21*m33, m11*m33-m31*m13, m21*m13-m11*m23,
                     m21*m32-m31*m22, m31*m12-m11*m32, m11*m22-m12*m21 )*d

    def scaling(s):
        """Return a scale transformation."""
        sx,sy,sz = s
        return mat3(sx, 0.0, 0.0,
                    0.0, sy, 0.0,
                    0.0, 0.0, sz)
    scaling = staticmethod(scaling)

    def rotation(angle, axis):
        """Return a rotation matrix."""
        axis = _vec3(axis)
        
        sqr_a = axis.x*axis.x
        sqr_b = axis.y*axis.y
        sqr_c = axis.z*axis.z
        len2  = sqr_a+sqr_b+sqr_c

        k2    = math.cos(angle)
        k1    = (1.0-k2)/len2
        k3    = math.sin(angle)/math.sqrt(len2)
        k1ab  = k1*axis.x*axis.y
        k1ac  = k1*axis.x*axis.z
        k1bc  = k1*axis.y*axis.z
        k3a   = k3*axis.x
        k3b   = k3*axis.y
        k3c   = k3*axis.z

        return mat3( k1*sqr_a+k2, k1ab-k3c, k1ac+k3b,
                     k1ab+k3c, k1*sqr_b+k2, k1bc-k3a,
                     k1ac-k3b, k1bc+k3a, k1*sqr_c+k2)
    rotation = staticmethod(rotation)

    def scale(self, s):
        sx = float(s[0])
        sy = float(s[1])
        sz = float(s[2])
        self.mlist[0] *= sx
        self.mlist[1] *= sy
        self.mlist[2] *= sz
        self.mlist[3] *= sx
        self.mlist[4] *= sy
        self.mlist[5] *= sz
        self.mlist[6] *= sx
        self.mlist[7] *= sy
        self.mlist[8] *= sz
        return self

    def rotate(self, angle, axis):
        R=self.rotation(angle, axis)
        self.mlist = (self*R).mlist
        return self

    def ortho(self):
        """Return a matrix with orthogonal base vectors.
        """

        m11,m12,m13,m21,m22,m23,m31,m32,m33 = self.mlist

        x = _vec3(m11, m21, m31)
        y = _vec3(m12, m22, m32)
        z = _vec3(m13, m23, m33)

        xl = x.length()
        xl*=xl
        y = y - ((x*y)/xl)*x
        z = z - ((x*z)/xl)*x

        yl = y.length()
        yl*=yl
        z = z - ((y*z)/yl)*y

        return mat3( x.x, y.x, z.x,
                     x.y, y.y, z.y,
                     x.z, y.z, z.z)

    def decompose(self):
        """Decomposes the matrix into a rotation and scaling part.

        Returns a tuple (rotation, scaling). The scaling part is given
        as a vec3, the rotation is still a mat3.
        """
        try:
            dummy = self.ortho()
        except ZeroDivisionError:
            return (mat3(1.0), _vec3(0))

        x = dummy.getColumn(0)
        y = dummy.getColumn(1)
        z = dummy.getColumn(2)
        xl = x.length()
        yl = y.length()
        zl = z.length()
        scale = _vec3(xl,yl,zl)
        
        x/=xl
        y/=yl
        z/=zl
        dummy.setColumn(0,x)
        dummy.setColumn(1,y)
        dummy.setColumn(2,z)
        if dummy.determinant()<0.0:
            dummy.setColumn(0,-x)
            scale.x=-scale.x

        return (dummy, scale)

    def fromEulerYXZ(x, y, z):
        """Initializes self from Euler angles."""
        A = math.cos(x)
        B = math.sin(x)
        C = math.cos(y)
        D = math.sin(y)
        E = math.cos(z)
        F = math.sin(z)
        CE = C*E
        CF = C*F
        DE = D*E
        DF = D*F

        return mat3( CE+DF*B, DE*B-CF, A*D,
                     A*F, A*E, -B,
                     CF*B-DE, DF+CE*B, A*C )

    fromEulerYXZ = staticmethod(fromEulerYXZ)

    def fromEulerZXY(x, y, z):
        """Initializes self from Euler angles."""
        A = math.cos(x)
        B = math.sin(x)
        C = math.cos(y)
        D = math.sin(y)
        E = math.cos(z)
        F = math.sin(z)
        CE = C*E
        CF = C*F
        DE = D*E
        DF = D*F

        return mat3( CE-DF*B, -A*F, DE+CF*B,
                     CF+DE*B, A*E, DF-CE*B,
                     -A*D, B, A*C )

    fromEulerZXY = staticmethod(fromEulerZXY)

    def fromEulerZYX(x, y, z):
        """Initializes self from Euler angles."""
        A = math.cos(x)
        B = math.sin(x)
        C = math.cos(y)
        D = math.sin(y)
        E = math.cos(z)
        F = math.sin(z)
        AE = A*E
        AF = A*F
        BE = B*E
        BF = B*F

        return mat3( C*E, BE*D-AF, AE*D+BF,
                     C*F, BF*D+AE, AF*D-BE,
                     -D, B*C, A*C )

    fromEulerZYX = staticmethod(fromEulerZYX)

    def fromEulerYZX(x, y, z):
        """Initializes self from Euler angles."""
        A = math.cos(x)
        B = math.sin(x)
        C = math.cos(y)
        D = math.sin(y)
        E = math.cos(z)
        F = math.sin(z)
        AC = A*C
        AD = A*D
        BC = B*C
        BD = B*D

        return mat3( C*E, BD-AC*F, BC*F+AD,
                     F, A*E, -B*E,
                     -D*E, AD*F+BC, AC-BD*F )

    fromEulerYZX = staticmethod(fromEulerYZX)

    def fromEulerXZY(x, y, z):
        """Initializes self from Euler angles."""
        A = math.cos(x)
        B = math.sin(x)
        C = math.cos(y)
        D = math.sin(y)
        E = math.cos(z)
        F = math.sin(z)
        AC = A*C
        AD = A*D
        BC = B*C
        BD = B*D

        return mat3( C*E, -F, D*E,
                     AC*F+BD, A*E, AD*F-BC,
                     BC*F-AD, B*E, BD*F+AC )

    fromEulerXZY = staticmethod(fromEulerXZY)

    def fromEulerXYZ(x, y, z):
        """Initializes self from Euler angles."""
        A = math.cos(x)
        B = math.sin(x)
        C = math.cos(y)
        D = math.sin(y)
        E = math.cos(z)
        F = math.sin(z)
        AE = A*E
        AF = A*F
        BE = B*E
        BF = B*F

        return mat3( C*E, -C*F, D,
                     AF+BE*D, AE-BF*D, -B*C,
                     BF-AE*D, BE+AF*D, A*C )

    fromEulerXYZ = staticmethod(fromEulerXYZ)

    def toEulerYXZ(self):
        """Return the Euler angles of a rotation matrix."""
        y,x,z = self._getRotation(2, True, True, True)
        return (x,y,z)

    def toEulerZXY(self):
        """Return the Euler angles of a rotation matrix."""
        z,x,y = self._getRotation(1, False, True, True)
        return (x,y,z)

    def toEulerZYX(self):
        """Return the Euler angles of a rotation matrix."""
        z,y,x = self._getRotation(0, True, True, True)
        return (x,y,z)

    def toEulerYZX(self):
        """Return the Euler angles of a rotation matrix."""
        y,z,x = self._getRotation(0, False, True, True)
        return (x,y,z)

    def toEulerXZY(self):
        """Return the Euler angles of a rotation matrix."""
        x,z,y = self._getRotation(1, True, True, True)
        return (x,y,z)

    def toEulerXYZ(self):
        """Return the Euler angles of a rotation matrix."""
        x,y,z = self._getRotation(2, False, True, True)
        return (x,y,z)

    def fromToRotation(_from, to):
        """Returns a rotation matrix that rotates one vector into another.

        The generated rotation matrix will rotate the vector _from into
        the vector to. _from and to must be unit vectors!

        This method is based on the code from:

        Tomas Möller, John Hughes
        Efficiently Building a Matrix to Rotate One Vector to Another
        Journal of Graphics Tools, 4(4):1-4, 1999
        http://www.acm.org/jgt/papers/MollerHughes99/
        """
        
        _from = _vec3(_from)
        to = _vec3(to)
        EPSILON = 0.000001
        e = _from*to
        f = abs(e)

        if (f>1.0-EPSILON):    # "from" and "to"-vector almost parallel
            # vector most nearly orthogonal to "from"
            fx = abs(_from.x)
            fy = abs(_from.y)
            fz = abs(_from.z)

            if (fx<fy):
                if (fx<fz):
                    x = _vec3(1.0, 0.0, 0.0)
                else:
                    x = _vec3(0.0, 0.0, 1.0)
            else:
                if (fy<fz):
                    x = _vec3(0.0, 1.0, 0.0)
                else:
                    x = _vec3(0.0, 0.0, 1.0)

            u = x-_from
            v = x-to

            c1 = 2.0/(u*u)
            c2 = 2.0/(v*v)
            c3 = c1*c2*u*v

            res = mat3()
            for i in range(3):
                for j in range(3):
                    res[i,j] =  - c1*u[i]*u[j] - c2*v[i]*v[j] + c3*v[i]*u[j]
                res[i,i] += 1.0
                
            return res
                
        else:  # the most common case, unless "from"="to", or "from"=-"to"
            v = _from.cross(to)
            h = 1.0/(1.0 + e)    # optimization by Gottfried Chen
            hvx = h*v.x
            hvz = h*v.z
            hvxy = hvx*v.y
            hvxz = hvx*v.z
            hvyz = hvz*v.y

            m11 = e + hvx*v.x
            m12 = hvxy - v.z
            m13 = hvxz + v.y

            m21 = hvxy + v.z
            m22 = e + h*v.y*v.y
            m23 = hvyz - v.x

            m31 = hvxz - v.y
            m32 = hvyz + v.x
            m33 = e + hvz*v.z

            return mat3(m11,m12,m13,m21,m22,m23,m31,m32,m33)

    fromToRotation = staticmethod(fromToRotation)


    def _getRotation(self, i, neg, alt, rev):
        """Get Euler angles in any of the 24 different conventions.
    
        The first four argument select a particular convention. The last three
        output arguments receive the angles. The order of the angles depends
        on the convention.
    
        See http://www.cgafaq.info/wiki/Euler_angles_from_matrix for the
        algorithm used.
    
        i: The index of the first axis (global rotations, s) or last axis (local rotations, r). 0=XZX, 1=YXY, 2=ZYZ
        neg: If true, the convention contains an odd permutation of the convention defined by i alone (i.e. the middle axis is replaced. For example, XZX -> XYX)
        alt: If true, the first and last axes are different. Local rotations: The first axis changes. For example, XZX -> YZX
        rev: If true, the first and last angle are exchanged. This toggles between global/local rotations. In all the concrete getRotation*() functions this is always true because all the functions assume local rotations.
        """    
        v = [self[0,i], self[1,i], self[2,i]]

        j,k,h = _eulerIndices(i, neg, alt)

        a = v[h]
        b = v[k]
        c,s,r = _eulerGivens(a, b)
        v[h] = r
        s1 = c*self[k,j] - s*self[h,j]
        c1 = c*self[k,k] - s*self[h,k]
        r1 = math.atan2(s1, c1)
        r2 = math.atan2(v[j], v[i])
        r3 = math.atan2(s, c)
        if alt:
            r3 = -r3
        if neg:
            r1 = -r1
            r2 = -r2
            r3 = -r3
        if rev:
            tmp = r1
            r1 = r3
            r3 = tmp
        return r1,r2,r3

def _eulerIndices(i, neg, alt):
    """Helper function for _getRotation()."""
    next = [1, 2, 0, 1]
    j = next[i+int(neg)]
    k = 3-i-j
    h = next[k+(1^int(neg)^int(alt))]
    return j,k,h

def _eulerGivens(a, b):
    """Helper function for _getRotation()."""
    global _epsilon
    
    absa = abs(a)
    absb = abs(b)
    # b=0?
    if absb<=_epsilon:
        if a>=0:
            c = 1.0
        else:
            c = -1.0
        return (c, 0.0, absa)
    # a=0?
    elif absa<=_epsilon:
        if b>=0:
            s = 1.0
        else:
            s = -1.0
        return (0.0, s, absb)
    # General case
    else:
        if absb>absa:
            t = a/b
            u = math.sqrt(1.0+t*t)
            if b<0:
                u = -u
            s = 1.0/u
            c = s*t
            r = b*u
        else:
            t = b/a
            u = math.sqrt(1.0+t*t)
            if (a<0):
                u = -u
            c = 1.0/u
            s = c*t
            r = a*u
        return c,s,r

######################################################################

if __name__=="__main__":

    vec3 = _vec3
    a=vec3(1,2,3)

    M = mat3("2,4,5,6")

    a=mat3(M)
    a[0,0]=17
    print (M)
    print (a)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: pluginmanager.py,v 1.2 2005/07/12 16:19:40 mbaas Exp $

## \file pluginmanager.py
## Contains the PluginManager class.

# (this line separates the \file comment from the module docs)

"""This module contains the plugin manager and associated functions.

The PluginManager class contains all the functionality to load plugins
and retrieve the objects inside the plugins.

The module already initializes a global plugin manager that can be
accessed using the functions with the same name than the manager methods.
"""

import os, os.path, sys, copy, types, glob, imp, inspect, traceback
try:
    import logging
except:
    # Define a dummy class that does nothing
    class DummyLogger:
        def debug(self, *args, **kwargs): pass
        def info(self, *args, **kwargs): pass
        def warning(self, *args, **kwargs): pass
        def error(self, *args, **kwargs): pass
        def critical(self, *args, **kwargs): pass
        def exception(self, *args, **kwargs): pass
    logging = DummyLogger()

# Exceptions:
class DuplicateObject(Exception):
    """Exception."""
    pass

class PluginAlreadyLoaded(Exception):
    """Exception."""
    pass

class UnknownFileType(Exception):
    """Exception."""
    pass

class ProtocolSpecsMissing(Exception):
    """Exception."""
    pass

class MissingName(Exception):
    """Exception."""
    pass


# Status flags
STATUS_OK         = 0x00
STATUS_EXCEPTION  = 0x01
STATUS_DUPLICATE  = 0x02

# PluginDescriptor
class PluginDescriptor:
    """Status descriptor of a plugin file.

    This class is just a container for a couple of info attributes about
    a plugin. The attributes are:

    - \b filename: The absolute file name of the plugin
    - \b filedesc: A file descriptor tuple (\em suffix, \em mode, \em type)
                 as returned by the get_suffixes() function in the \c imp
                 module (\em type is one of imp.PY_SOURCE, imp.PY_COMPILED
                 or imp.C_EXTENSION)
    - \b objcount: Number of imported objects (usually this should be
                 #classes + #functions)
    - \b classcount: Number of imported classes
    - \b funccount: Number of imported functions
    - \b status: Status flags. If everything is ok the status is STATUS_OK,
                otherwise its a combination of STATUS_EXCEPTION and
                STATUS_DUPLICATE.
    - \b traceback: A string containing the traceback message if the 
                STATUS_EXCEPTION flag is set.
    - \b objdescs: A list of object descriptors of all imported objects
    - \b module: The module object of the imported plugin
    - \b modulename: The name of the plugin module

    \see PluginManager, PluginObjectDescriptor
    """
    def __init__(self, filename="<unknown>", filedesc=None,
                 status=STATUS_OK, traceback=None,
                 objcount=0, classcount=0, funccount=0, module=None):
        self.filename   = filename
        self.filedesc   = filedesc
        self.status     = status
        self.traceback  = traceback
        self.objcount   = objcount
        self.classcount = classcount
        self.funccount  = funccount
        self.module     = module
        self.objdescs   = []
        self.modulename = None


    def __repr__(self):
        res = '<PluginDescriptor file:"%s" status:%s #objs:%s #cls:%s #func:%s>'%(self.filename, self.status, self.objcount, self.classcount, self.funccount)
        return res

    def __str__(self):
        res = 'Plugin "%s":\n'%self.filename
        s = "?"
        if self.filedesc!=None:
            t = self.filedesc[2]
            if t==imp.PY_SOURCE:
                s = "Python source"
            elif t==imp.PY_COMPILED:
                s = "Compiled Python code"
            elif t==imp.C_EXTENSION:
                s = "C extension"
            else:
                s += " (%s)"%t
        res += "  File type  : %s\n"%s
        res += "  Status     : %s\n"%self.status
        res += "  #Objects   : %s\n"%self.objcount
        res += "  #Classes   : %s\n"%self.classcount
        res += "  #Functions : %s\n"%self.funccount
        res += "  Objects    : %s\n"%self.objdescs
        res += "  Module     : %s"%getattr(self.module, "__name__", None)
        if self.status==STATUS_EXCEPTION:
            res += "\n  Traceback:\n%s"%self.traceback
        return res

    def setStatusFlag(self, flags):
        self.status |= flags

# PluginObjectDescriptor
class PluginObjectDescriptor:
    """Descriptor for a plugin object.

    - \b object: The actual plugin object
    - \b name: Object name
    - \b plugindesc: Associated plugin
    - \b status: Status flags. If everything is ok the status is STATUS_OK,
                otherwise its STATUS_DUPLICATE.

    \see PluginManager, PluginDescriptor
    """
    def __init__(self, object=None, name=None,
                 plugindesc=None, status=STATUS_OK):
        self.object = object
        self.name = name
        self.plugindesc = plugindesc
        self.status = status

    def __repr__(self):
        res = "<Plugin object '%s', status=%s>"%(self.objectIdentifier(), self.status)
        return res

    def moduleName(self):
        if self.plugindesc==None:
            return None
        else:
            return self.plugindesc.modulename

    def objectIdentifier(self):
        res = self.name
        modname = self.moduleName()
        if modname!=None:
            res = modname+"."+res
        return res

    def setStatusFlag(self, flags):
        self.status |= flags


# PluginManager
class PluginManager:
    """Loads and manages plugins.

    This class imports and manages plugins. A plugin is a file that
    gets imported just like a regular Python module. Each plugin can
    be uniquely identified either by its absolute filename or its
    module name. The module name is only available once the plugin is
    loaded. Any objects (usually classes) that are specially marked
    will be made available by the plugin manager for later retrieval.

    A class gets imported if it has an attribute \c _protocols which
    is a list of supported protocols (a protocol is identified by an
    arbitrary hashable object). Each class can be uniquely identified
    by its module name and class name.
    
    Plugins are loaded via calls to importPlugin() or importPlugins().
    You can also register any class as plugin class any time via
    the register() function. When a plugin is loaded or an object
    is registered the plugin/object is represented by a descriptor
    class (PluginDescriptor resp. PluginObjectDescriptor). These
    descriptor objects contain information about the plugin/object.
    The object descriptor also holds a reference to the actual plugin
    object.

    You can iterate over all currently available protocols via the
    iterProtocols() method. The iterProtoObjects() method can be used
    to iterate over all objects that support a particular protocol
    and iterObjects() can be used to iterate over all objects.
    The default iterator iterates over all plugin descriptors.

    Example plugin class:
    
    \code
    # file: plugin.py
    
    class PluginClass:

        _protocols = ["MyProtocol"]

        def __init__(self):
            ...
        ...
    \endcode

    And here's how to use it:

    \code
    pm = PluginManager()
    
    # Load the plugin
    pdesc = pm.importPlugin("plugin.py")
    if pdesc.status!=STATUS_OK:
      # there was an error...

    # Search a plugin class
    objdesc = pm.findObject("plugin.PluginClass")
    # Create an instance of the plugin class...
    PluginClass = objdesc.object
    instance = PluginClass()
    
    \endcode

    \see PluginDescriptor, PluginObjectDescriptor
    """

    def __init__(self):
        """Constructor."""
        
        # A list with plugin descriptors of all imported plugins
        self._plugins = []

        # A dictionary containing all objects.
        # Key is the object identifier string, value is the descriptor
        self._objects = {}

        # A dictionary with all available protocols. The key is the protocol
        # identifier (an arbitrary hashable object), the value is a list
        # of object descriptors.
        self._protocols = {}
        


    ######################################################################
    # Neue Schnittstelle:

    def __iter__(self):
        """Return an iterator that iterates over all plugin descriptors."""
        return iter(self._plugins)

    def iterProtocols(self):
        """Return an iterator that iterates over all protocols."""
        return iter(self._protocols)

    def iterProtoObjects(self, proto):
        """Return an iterator that iterates over all object descriptors supporting a particular protocol."""
        if proto in self._protocols:
            return iter(self._protocols[proto])
        else:
            return iter([])

    def iterObjects(self):
        """Return an iterator that iterates over all object descriptors.
        """
        return iter(list(self._objects.values()))

    # removeAll
    def removeAll(self):
        """Remove all plugins and plugin objects.
        """
        
        while len(self._objects)>0:
            key = list(self._objects.keys())[0]
            self.remove(self._objects[key])
            
        self._plugins = []
        self._objects = {}
        self._protocols = {}

    # importPlugin
    def importPlugin(self, filename, overwrite=False):
        """Load a plugin file.

        The given file is executed, i.e. it is imported like a module.
        Any object (usually class or function) that has an attribute
        "_protocols" is stored in the plugin manager. This attribute
        is a list of protocol specifiers (which are just arbitrary
        hashable objects).

        The file must have a suffix that's recognized as a module by
        the Python interpreter (one of the suffixes returned by
        imp.get_suffixes()), otherwise an UnknownFileType exception is
        thrown.

        It is possible to load a plugin several times if \a overwrite
        is set to True. The new definitions override the previous ones.
        However, if references or instances to old objects are kept somewhere
        they still refer to the old definition. When writing a plugin you
        should always bear in mind that the file could be executed several
        times and write your initialization code accordingly.

        The function returns a PluginDescriptor object which contains
        information about the imported plugin. 
        
        \param filename (\c str) Name of the plugin file (including path)
        \param overwrite (\c bool) Flag that indicates if objects are allowed
                  to overwrite existing plugin objects
        \return Plugin descriptor object (\c PluginDescriptor)
        \todo Lokalen Modul-Import innerhalb des Plugins testen
        """
        
        # Make the filename absolute (and normalized)
        filename = os.path.abspath(filename)
        # path: Absolute path to the file
        path, name = os.path.split(filename)
        # ext: Extension of the file
        name, ext  = os.path.splitext(filename)
        # modname: Module name (only the file name, without path and extension)
        modname = os.path.basename(name)

        # Check if the file is already imported
        oldpdesc = self.findPlugin(filename)
        if oldpdesc!=None:
            if overwrite:
                self.removePlugin(oldpdesc)
            else:
                raise PluginAlreadyLoaded("Plugin '%s' is already loaded."%filename)

        # Find the file description tuple  (suffix, mode, type)
        filedesc = [x for x in imp.get_suffixes() if x[0]==ext]
        if filedesc==[]:
            raise UnknownFileType('File "%s" is of unknown type."'%filename)
        filedesc = filedesc[0]

        # Create a plugin descriptor class where all sorts of
        # info is gathered
        pdesc = PluginDescriptor(filename=filename, filedesc=filedesc)

        # Open the file
        f = file(filename)

        # Add the plugin to the list of imported plugins
        self._plugins.append(pdesc)

        # Import the file as a module
        # The path to the file is added to the search path for modules,
        # so the file can import local modules
        try:
            oldpath = copy.copy(sys.path)
            sys.path.append(path)
            try:
                mod = imp.load_module(modname, f, filename, filedesc)
                pdesc.module = mod
                pdesc.modulename = mod.__name__
            finally:
                f.close()
                sys.path = oldpath
        except:
            # Extract the traceback message...
            lst = traceback.extract_tb(sys.exc_info()[2])
            flst = traceback.format_list(lst[1:])
            pdesc.traceback = "".join(flst)
            # Add the actual exception to the message...
            t,v = sys.exc_info()[:2]
            elst = traceback.format_exception_only(t,v)
            pdesc.traceback += "".join(elst)
            
            pdesc.setStatusFlag(STATUS_EXCEPTION)
            return pdesc

        # Examine the module namespace and register all objects that
        # have a "_protocols" attribute...
        for objname in mod.__dict__:
            obj = mod.__dict__[objname]
            if hasattr(obj, "_protocols"):
                self.register(obj, pdesc=pdesc)
                
        return pdesc

    # importPlugins
    def importPlugins(self, plugins, out = sys.stderr):
        """Import several plugins at once.

        \a plugins can be a single file/directory name or a sequence
        of file/directory names. Directories are recursively descended.

        \return A list of plugin descriptors.
        """

        if isinstance(plugins, str):
            plugins = [plugins]
        
        res = []
        for plugin in plugins:
            # Check if the file/dir is there...
            if not os.path.exists(plugin):
                out.write('ERROR: Plugin file or directory "%s" does not exist.\n'%plugin)
                continue

            # Get a list of files (either the single file or the contents
            # of the directory)
            if os.path.isfile(plugin):
                files = [plugin]
            else:
                files = glob.glob(os.path.join(plugin,"*"))

            # Import each file...
            for f in files:
                # Directory? then import recursively
                if os.path.isdir(f):
                    res += self.importPlugins(f, out)
                # File...
                else:
                    name, ext = os.path.splitext(f)
                    ext = ext.lower()
                    # Only try to import if the file is actually a
                    # Python module (other than .pyc)...
                    filedesc = [x for x in imp.get_suffixes() if x[0]==ext]
                    if filedesc!=[] and ext!=".pyc":
                        out.write('Loading plugin "'+f+'"...\n')
                        d = self.importPlugin(f)
                        res.append(d)
        return res

    # register
    def register(self, obj, name=None, pdesc=None, overwrite=False):
        """Register a plugin object.

        If \a name is None then the name of the object will be used.      

        \param obj  The plugin object
        \param name (\c str) The name of the object or None
        \param pdesc (\c PluginDescriptor) The corresponding plugin or None
        \param overwrite (\c bool) If True the object will overwrite any existing objects.
        \return Object descriptor (\c PluginObjectDescriptor)
        """

        # Fill in name automatically?
        if name==None:
            name = getattr(obj, "__name__", None)
            if name==None:
                raise MissingName("Cannot determine the name of the object.")
        if not hasattr(obj, "_protocols"):
            raise ProtocolSpecsMissing("Attribut '_protocols' in object '%s' is missing."%name)

        # Create an object descriptor
        desc = PluginObjectDescriptor(object=obj, name=name, plugindesc=pdesc)

        # Check if the object already exists...
        id = desc.objectIdentifier()
        if id in self._objects:
            if overwrite:
                olddesc = self._objects[id]
                self._removeObjectDesc(olddesc)
            else:                
                raise DuplicateObject("Plugin object '%s' already exists."%id)

        # Add the descriptor to the protocol lists...
        self._insertObjectDesc(desc)

        return desc

    # remove
    def remove(self, objdesc):
        """Remove a plugin object.

        \param objdesc (\c PluginObjectDescriptor) Object descriptor
        """
        # Check if the object is really managed by the plugin manager
        id = objdesc.objectIdentifier()
        if objdesc.objectIdentifier() not in self._objects:
            raise ValueError("The object '%s' does not exist."%id)

        self._removeObjectDesc(objdesc)

    # removePlugin
    def removePlugin(self, pdesc):
        """Remove a plugin and its imported objects from the manager.

        Note: If some classes are still used somewhere those classes
        remain active. It's just that you won't be able to retrieve
        those classes via the plugin manager.

        \param pdesc (\c PluginDescriptor) Descriptor of the plugin that should be removed
        """
        # Remove all imported objects...
        while len(pdesc.objdescs)>0:
            self.remove(pdesc.objdescs[0])

        # Remove the plugin
        self._plugins.remove(pdesc)


    # findPlugin
    def findPlugin(self, filename):
        """Return the plugin descriptor for a given filename.

        \param filename (\c str) Plugin filename
        \return Plugin descriptor (\c PluginDescriptor) or None
        """
        filename = os.path.abspath(filename)
        pl = [d for d in self._plugins if d.filename==filename]
        if pl==[]:
            return None
        else:
            return pl[0]

    # findObject
    def findObject(self, name, modname=None):
        """Return the object descriptor for the specified object.

        \param name (\c str) Object name
        \param modname (\c str) Module name
        \return Object descriptor or None
        """
        if modname!=None:
            name = modname+"."+name
        return self._objects.get(name, None)

        
    ######################################################################
    ## protected:

    def _insertObjectDesc(self, desc):
        """Insert a plugin object.

        \pre The object isn't already stored in the manager
        \pre The field desc.name contains a string
        \pre The object has a valid _protocols attribute (a sequence)
        \param desc (\c PluginObjectDescriptor) Object descriptor
        """

        # Add the object descriptor to the manager
        self._objects[desc.objectIdentifier()] = desc

        # Insert the object descriptor into the corresponding protocol dicts
        for prot in desc.object._protocols:
            if prot not in self._protocols:
                self._protocols[prot] = []
            self._protocols[prot].append(desc)

        # Update the plugin descriptor
        pdesc = desc.plugindesc
        if pdesc!=None:
            pdesc.objdescs.append(desc)
            pdesc.objcount += 1
            if inspect.isclass(desc.object):
                pdesc.classcount += 1
            if inspect.isfunction(desc.object):
                pdesc.funccount += 1


    def _removeObjectDesc(self, desc):
        """Remove a plugin object from the manager.

        \pre The object is stored in the manager
        \param desc (\c PluginObjectDescriptor) Object descriptor
        """

        # Update the plugin descriptor
        pdesc = desc.plugindesc
        if pdesc!=None:
            pdesc.objdescs.remove(desc)
            pdesc.objcount -= 1
            if inspect.isclass(desc.object):
                pdesc.classcount -= 1
            if inspect.isfunction(desc.object):
                pdesc.funccount -= 1
        
        # Remove the object from the protocol dicts
        for prot in desc.object._protocols:
            self._protocols[prot].remove(desc)
                
        # Remove the object descriptor
        del self._objects[desc.objectIdentifier()]
        

######################################################################

# Global default plugin manager
_plugin_manager = PluginManager()

def iterPlugins():
    """Iterate over all global plugins.

    This is equivalent to calling \c iter(pm) (with \c pm being the global
    plugin manager).

    \see PluginManager::__iter__()
    """
    global _plugin_manager
    return iter(_plugin_manager)

def iterProtocols():
    """Global %iterPrototols() function.

    \see PluginManager::iterProtocols()
    """
    global _plugin_manager
    return _plugin_manager.iterProtocols()

def iterProtoObjects(proto):
    """Global %iterProtoObjects() function.

    \see PluginManager::iterProtoObjects()
    """
    global _plugin_manager
    return _plugin_manager.iterProtoObjects(proto)

def iterObjects(proto):
    """Global %iterObjects() function.

    \see PluginManager::iterObjects()
    """
    global _plugin_manager
    return _plugin_manager.iterObjects()

def removeAll(obj):
    """Global %removeAll() function.

    \see PluginManager::removeAll()
    """
    global _plugin_manager
    _plugin_manager.removeAll()

def importPlugin(filename, overwrite=False):
    """Global %importPlugin() function.

    \see PluginManager::importPlugin()
    """
    global _plugin_manager
    return _plugin_manager.importPlugin(filename, overwrite)

def importPlugins(dir, out=sys.stdout):
    """Global %importPlugins() function.

    \see PluginManager::importPlugins()
    """
    global _plugin_manager
    return _plugin_manager.importPlugins(dir, out)

def register(obj, name=None, pdesc=None, overwrite=False):
    """Global %register() function.

    \see PluginManager::register()
    """
    global _plugin_manager
    return _plugin_manager.register(obj, name, pdesc, overwrite)

def remove(objdesc):
    """Global %remove() function.

    \see PluginManager::remove()
    """
    global _plugin_manager
    _plugin_manager.remove(objdesc)

def removePlugin(pdesc):
    """Global %removePlugin() function.

    \see PluginManager::removePlugin()
    """
    global _plugin_manager
    _plugin_manager.removePlugin(pdesc)

def findPlugin(filename):
    """Global %findPlugin() function.

    \see PluginManager::findPlugin()
    """
    global _plugin_manager
    return _plugin_manager.findPlugin(filename)

def findObject(name, modname=None):
    """Global %findObject() function.

    \see PluginManager::findObject()
    """
    global _plugin_manager
    return _plugin_manager.findObject(name, modname)



######################################################################

if __name__=="__main__":

    class TestClass:
        _protocols = ["myproto"]

    pm = PluginManager()

#    desc = pm.register(TestClass)                      
#    print desc.objectIdentifier()

    pm.importPlugins(".")
#    pm.importPlugin("testplugin.py")
#    pm.importPlugin("testplugin.py", overwrite=True)

    for pdesc in pm:
        print (pdesc)

    for prot in pm.iterProtocols():
        print(("Protokoll: %s"%prot))
        for odesc in pm.iterProtoObjects(prot):
            print (odesc)

    print ("Objects:")
    for desc in pm.iterObjects():
        print (desc)

#    print "Plugins:",pm._plugins
#    print "Objects:",pm._objects
#    print "Protocols:",pm._protocols




# Render a RIB file

import os, subprocess
import cgkit.jobqueue

class renderrib(cgkit.jobqueue.JobProc):
    
    def __init__(self, rib, renderer="aqsis"):
        cgkit.jobqueue.JobProc.__init__(self)
        self._rib = rib
        self._renderer = renderer.lower()
        
    def run(self):
        args = [self._rendererExecutable(), self._rib]
        cmd = " ".join(args)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                shell=True)
        outdata,errdata = proc.communicate()
        outdata = outdata.rstrip()
        errdata = errdata.rstrip()
        print("OUT")
        print(outdata)
        print("ERR")
        print(errdata)
        print("RET")
        print((proc.returncode))
        
        if proc.returncode!=0 or errdata!="":
            self.setError()
        
        
    def _rendererExecutable(self):
        """Return the name of the renderer executable.
        """
        if self._renderer=="aqsis":
            return "aqsis"
        elif self._renderer=="pixie":
            return "rndr"
        elif self._renderer=="3delight":
            return "renderdl"
        elif self._renderer=="prman":
            return "prman"
        else:
            raise ValueError("Unknown renderer: %s"%self._renderer)
